from cosymlib.molecule.geometry import Geometry
from cosymlib.molecule.electronic_structure import ElectronicStructure
from cosymlib.symmetry.wfnsym import Wfnsym
from cosymlib.symmetry.pointgroup import CalculatePointGroup
from cosymlib.simulation import ExtendedHuckel


class Molecule:

    def __init__(self, geometry, ee=None):

        if not geometry:
            print('No geometry found in the input file, check out input file for possible errors')
            exit()
        self._geometry = geometry
        self._name = geometry.get_name()
        self._electronic_structure = ee
        if ee is not None:
            self._wfnsym = Wfnsym(self)

    def get_name(self):
        return self._name

    # def set_wfnsym(self):
        # self._electronic_structure = ee


    @property
    def geometry(self):
        return self._geometry

    @property
    def electronic_structure(self):
        if self._electronic_structure is None:
            print('No electronic structure found in the input file. Starting a extended-huckel calculation to determine'
                  'the molecular orbital coefficients...')
            eh = ExtendedHuckel(self.geometry)
            self._electronic_structure = ElectronicStructure(basis=eh.get_basis(),
                                                             orbital_coefficients=[eh.get_mo_coefficients(), []],
                                                             mo_energies=eh.get_mo_energies(),
                                                             valence_only=True)
        return self._electronic_structure

    def get_mo_symmetry(self, group, vector_axis1=None, vector_axis2=None, center=None):
        from copy import deepcopy
        wfnsym = deepcopy(Wfnsym(self))
        return wfnsym.results(group, vector_axis1, vector_axis2, center)

    def get_pointgroup(self, tol=0.01):
        return CalculatePointGroup(self._geometry, tolerance=tol).get_point_group()



import itertools
import numpy as np
from cosymlib.symmetry.pointgroup.operations import inversion, rotation_matrix, reflection
from cosymlib.symmetry.pointgroup import sym_tools


class CalculatePointGroup:

    def __init__(self, geometry, tolerance=0.01, tol=0.3):
        self.tolerance = tolerance
        self._tol = tol
        self._symbols = geometry.get_symbols()
        self.schoenflies_symbol = ''
        self.sym_op = []
        self.rot_sym = []
        self._max_order = ''
        self._cent_coord = geometry.get_positions() - sym_tools.get_center_mass(self._symbols, geometry.get_positions())
        self.set_eigen_values_vectors()

    def get_schoenflies_symbol(self):
        if self.schoenflies_symbol == '':
            val = self._eigenvalues
            if any([i < self.tolerance for i in val]):
                # print('Linear molecule found')
                self._lineal()
            elif abs(val[0] - val[1]) < self.tolerance and abs(val[0] - val[2]) < self.tolerance:
                # print('Spherical molecule found')
                self._spherical()
            elif abs(val[0] - val[1]) > self.tolerance and abs(val[1] - val[2]) > self.tolerance \
                    and abs(val[0] - val[2]) > self.tolerance:
                # print('Asymmetric molecule found')
                self._asymmetric()
            else:
                # print('Symmetric molecule found')
                self._symmetric()

        return self.schoenflies_symbol

    # def which_schoenflies_symbol(self):
    #     print('The point group of the molecule is: ', self.get_schoenflies_symbol())

    def set_eigen_values_vectors(self):
        I = sym_tools.get_inertia_tensor(self._symbols, self._cent_coord)
        eigenvalues, eigenvectors = np.linalg.eigh(I)
        self._cent_coord = np.dot(self._cent_coord, eigenvectors)
        eigenvectors = np.dot(eigenvectors, eigenvectors.T)
        # I = sym_tools.get_inertia_tensor(self._symbols, self._cent_coord)
        # eigenvalues, eigenvectors = np.linalg.eigh(I)
        self._eigenvalues = eigenvalues
        self._eigenvectors = eigenvectors.T

    def _lineal(self):

        if self.check_op(inversion()):
            self.schoenflies_symbol = 'Dinfh'
        else:
            self.schoenflies_symbol = 'Cinfh'

    def _asymmetric(self):

        for axis in self._eigenvectors:
            c2 = rotation_matrix(axis, 180)
            if self.check_op(c2):
                self.sym_op.append(c2)
                self.rot_sym.append((axis, 2))

        if len(self.sym_op) == 0:
            self._no_rot_axis()
        elif len(self.sym_op) == 1:
            self._cyclic()
        else:
            self._dihedral()

    def _symmetric(self):

        if abs(self._eigenvalues[0] - self._eigenvalues[1]) < self.tolerance:
            idx = 2
        elif abs(self._eigenvalues[0] - self._eigenvalues[2]) < self.tolerance:
            idx = 1
        else:
            idx = 0

        main_axis = self._eigenvectors[idx, :]
        self.check_order_rot(main_axis)
        self.rot_sym.append((main_axis, self._max_order))

        # Obtain all the c2 vectors perpendicular to main axis
        for ida, axis in enumerate(self._eigenvectors):
            if ida != idx:
                p_axis = axis
                break
        for angle in np.linspace(0, 360, 144, endpoint=False):
            axis = np.matmul(np.array(rotation_matrix(main_axis, angle)), np.array(p_axis))
            c2 = rotation_matrix(axis, 360 / 2)
            if self.check_op(c2):
                self.sym_op.append(c2)
                self.rot_sym.append((axis / np.linalg.norm(axis), 2))

        self.check_consistency()
        if len(self.rot_sym) >= 2:
            self._dihedral()
        elif len(self.rot_sym) == 1:
            self._cyclic()
        else:
            self._no_rot_axis()

    def _spherical(self):

        # I or Ih
        for axis in self._eigenvectors:
            c5 = rotation_matrix(axis, 360 / 5)
            if self.check_op(c5):
                self.sym_op.append(c5)
                self.rot_sym.append((axis, 5))
                self.schoenflies_symbol = "I"
                self._max_order = 5
        if self.schoenflies_symbol and self.check_op(inversion()):
            self.schoenflies_symbol += 'h'

        # O or Oh
        if not self.schoenflies_symbol:
            for axis in self._eigenvectors:
                c4 = rotation_matrix(axis, 360 / 4)
                if self.check_op(c4):
                    self.sym_op.append(c4)
                    self.rot_sym.append((axis, 4))
                    self.schoenflies_symbol = "O"
                    self._max_order = 4
            if self.schoenflies_symbol and self.check_op(inversion()):
                self.schoenflies_symbol += 'h'

        # Obtain all the c2 vectors and matrices
        for angle1 in np.linspace(0, 360, 72, endpoint=False):
            axis1 = np.matmul(np.array(rotation_matrix([1, 0, 0], angle1)), np.array([0, 0, 1]))
            for angle2 in np.linspace(0, 180, 72, endpoint=False):
                axis2 = np.matmul(np.array(rotation_matrix([0, 1, 0], angle2)), axis1)
                c2 = rotation_matrix(axis2, 360 / 2)
                if self.check_op(c2):
                    self.sym_op.append(c2)
                    self.rot_sym.append((axis2 / np.linalg.norm(axis2), 2))

        if self.rot_sym:
            self.check_consistency()

        # T or Td, Th
        m_type = ''
        if not self.schoenflies_symbol:
            # Obtain all the c3 vectors and matrices
            for angle1 in np.linspace(0, 360, 72, endpoint=False):
                axis1 = np.matmul(np.array(rotation_matrix([1, 0, 0], angle1)), np.array([0, 0, 1]))
                for angle2 in np.linspace(0, 180, 72, endpoint=False):
                    axis2 = np.matmul(np.array(rotation_matrix([0, 1, 0], angle2)), axis1)
                    c3 = rotation_matrix(axis2, 360 / 3)
                    if self.check_op(c3):
                        self.sym_op.append(c3)
                        self.rot_sym.append((axis2 / np.linalg.norm(axis2), 3))
            self.schoenflies_symbol = "T"
            self._max_order = 3
            self.check_consistency()
            for axis, order in self.rot_sym:
                if order == 3:
                    m_type = self.mirror_plane(axis)
                    break
            if m_type == '':
                if self.check_op(inversion()):
                    self.rot_sym.append(([0, 0, 0], 'i'))
                    self.sym_op.append(inversion())
                    self.schoenflies_symbol += 'h'
                # else:
                #     self.schoenflies_symbol += 'd'
            else:
                self.schoenflies_symbol += 'd'
            self._improper_rot()

    def _no_rot_axis(self):
        for vector in self._eigenvectors:
            if self.check_op(reflection(vector)):
                self.schoenflies_symbol = 'Cs'
                self.rot_sym.append((vector, 'Sh'))
                self.sym_op.append(reflection(vector))
                break
            else:
                if self.check_op(inversion()):
                    self.schoenflies_symbol = 'Ci'
                    self.rot_sym.append((vector, 'i'))
                    self.sym_op.append(inversion())
                else:
                    self.schoenflies_symbol = 'C1'

    def _cyclic(self):
        main_axis, order = max(self.rot_sym, key=lambda v: v[1])
        self.schoenflies_symbol = "C{}".format(order)
        m_type = self.mirror_plane(main_axis)
        if m_type != '':
            self.schoenflies_symbol += m_type
        else:
            self._improper_rot()
            for vector, order in self.rot_sym:
                if 'S' in str(order):
                    self.schoenflies_symbol = order

    def _dihedral(self):
        main_axis, order = max(self.rot_sym, key=lambda v: v[1])
        self.schoenflies_symbol = "D{}".format(order)
        m_type = self.mirror_plane(main_axis)
        if m_type != '':
            self.schoenflies_symbol += m_type
        self.check_consistency()

    def mirror_plane(self, main_axis):

        m_type = ''
        if self.check_op(reflection(main_axis)):
            self.rot_sym.append((main_axis, 'Sh'))
            self.sym_op.append(reflection(main_axis))
            m_type = 'h'
        else:
            # Sv planes #
            p_axis = np.random.randn(3)
            p_axis -= np.dot(p_axis, main_axis)*main_axis
            p_axis /= np.linalg.norm(p_axis)
            for angle in np.linspace(0, 360, 144, endpoint=False):
                axis = np.matmul(np.array(rotation_matrix(main_axis, angle)), np.array(p_axis))
                possible_mirror = reflection(axis)
                if self.check_op(possible_mirror):
                    m_type = 'v'
                    self.rot_sym.append((axis, 'Sv'))
                    self.sym_op.append(possible_mirror)

            # Sd planes #
            c2_axes, gt_axes = [], []
            for axis, order in self.rot_sym:
                if order == 2:
                    c2_axes.append(axis)
                if order == self._max_order:
                    gt_axes.append(axis)

            for c2_1, c2_2 in itertools.combinations(c2_axes, 2):
                c2_vector = c2_1 + c2_2
                if all(c2_vector < 1e-8):
                    continue
                else:
                    c2_vector /= np.linalg.norm(c2_vector)
                    for gt_vector in gt_axes:
                        p_axis = np.cross(c2_vector, gt_vector)
                        possible_mirror = reflection(p_axis)
                        if self.check_op(reflection(p_axis)):
                            m_type = 'd'
                            self.rot_sym.append((p_axis / np.linalg.norm(p_axis), 'Sd'))
                            self.sym_op.append(possible_mirror)

        return m_type

    def _improper_rot(self):

        for axis, order in self.rot_sym:
            if order == 2:
                ref_mat = reflection(axis)
                for n in range(20, 0, -2):
                    Cn = rotation_matrix(axis, 360 / n)
                    Sn = np.dot(Cn, ref_mat)
                    if self.check_op(Sn):
                        self.rot_sym.append((axis, 'S' + str(n)))
                        self.sym_op.append(Sn)
                        break

    def check_order_rot(self, axis):

        order = 0
        for i in list(range(2, 9)):
            Cn = rotation_matrix(axis, 360 / i)
            if self.check_op(Cn):
                order = i
                self.sym_op.append(Cn)
        self._max_order = order

    def check_consistency(self):

        non_repeated_vectors, matrix_op = [], []
        non_repeated_vectors.append(self.rot_sym[0])
        matrix_op.append(self.sym_op[0])
        for idx, element in enumerate(self.rot_sym):
            non_repeated = True
            for vector, order in non_repeated_vectors:
                diff = abs(element[0] - vector)
                add = abs(element[0] + vector)
                if (all(diff < self._tol) or all(add < self._tol)) \
                        and element[1] == order:
                    non_repeated = False
            if non_repeated:
                non_repeated_vectors.append((element[0], element[1]))
                matrix_op.append(self.sym_op[idx])

        self.rot_sym = non_repeated_vectors
        self.sym_op = matrix_op

    def check_op(self, sym_matrix, tol=1e-02):

        def find_atom_in_list(atom, tol):
            atom_ind = []
            difference = [np.sqrt(np.sum(x)) for x in np.square(self._cent_coord - atom)]
            for idx, dif in enumerate(difference < np.sqrt(tol)):
                if dif:
                    atom_ind.append(idx)
            return atom_ind

        for idx, atom in enumerate(self._cent_coord):
            atom_op = np.dot(sym_matrix, atom)
            ind = find_atom_in_list(atom_op, tol)
            if not (len(ind) == 1 and self._symbols[ind[0]] == self._symbols[idx]):
                return False
        return True

    def get_point_group(self):
        return self.get_schoenflies_symbol()

    def get_rot_vector(self):
        self.rot_sym.insert(0, ([0, 0, 0], 'E'))
        return self.rot_sym

    def get_sym_matrix(self):
        self.sym_op.insert(0, np.identity(3))
        return self.sym_op

    def get_centerd_coordinates(self):
        return self._cent_coord



import numpy as np


def build_symmetry_operated_matrices(group, molecule, parsed_data):

    geometry = molecule.geometry
    sym_lables = parsed_data.SymLab
    sym_matrices = parsed_data.SymMat

    sep_line = '--------------------------------------------\n'
    txt_sym = 'MEASURES OF THE SYMMETRY GROUP:   {}\n'.format(group)
    txt_sym += 'Basis: {}\n'.format(list(molecule.electronic_structure.basis.keys())[0])
    txt_sym += sep_line
    txt_sym += ' Atomic Coordinates (Angstroms)\n'
    txt_sym += sep_line
    for idn, array in enumerate(geometry.get_positions()):
        txt_sym += '{:2} {:11.6f} {:11.6f} {:11.6f}\n'.format(geometry.get_symbols()[idn],
                                                              array[0], array[1], array[2])
    txt_sym += sep_line
    for i, group in enumerate(sym_lables):
        txt_sym += '\n'
        txt_sym += '@@@ Operation {0}: {1}'.format(i + 1, group)
        txt_sym += '\nSymmetry Transformation matrix\n'
        for array in sym_matrices[i]:
            txt_sym += ' {:11.6f} {:11.6f} {:11.6f}\n'.format(array[0], array[1], array[2])
        txt_sym += '\n'
        txt_sym += 'Symmetry Transformed Atomic Coordinates (Angstroms)\n'

        for idn, array in enumerate(geometry.get_positions()):
            array2 = np.dot(array, sym_matrices[i].T)
            txt_sym += '{:2} {:11.6f} {:11.6f} {:11.6f}\n'.format(geometry.get_symbols()[idn],
                                                                  array2[0], array2[1], array2[2])

    return txt_sym


def build_symmetry_overlap_analysis(parsed_data):

    ir_labels = parsed_data.SymLab
    ideal_gt = parsed_data.ideal_gt
    mo_soevs_a = parsed_data.mo_SOEVs_a
    mo_soevs_b = parsed_data.mo_SOEVs_b
    wf_soevs_a = parsed_data.wf_SOEVs_a
    wf_soevs_b = parsed_data.wf_SOEVs_b
    wf_soevs = parsed_data.wf_SOEVs
    grim_coef = parsed_data.grim_coef
    csm_coef = parsed_data.csm_coef

    sep_line = '   -------------------------------------------------------------------------------' \
               '------------------------------------------------------------------------\n'

    txt_sym = '\nIdeal Group Table\n'
    txt_sym += sep_line
    txt_sym += '     ' + '  '.join(['{:^7}'.format(s) for s in ir_labels])
    txt_sym += '\n'
    txt_sym += sep_line
    for i, line in enumerate(ideal_gt):
        txt_sym += '{:4}'.format(ir_labels[i]) + '  '.join(['{:7.3f}'.format(s) for s in line])
        txt_sym += '\n'
    txt_sym += sep_line

    txt_sym += '\nAlpha MOs: Symmetry Overlap Expectation Values\n'
    txt_sym += sep_line
    txt_sym += '     ' + '  '.join(['{:^7}'.format(s) for s in ir_labels])
    txt_sym += '\n'
    txt_sym += sep_line

    for i, line in enumerate(mo_soevs_a):
        txt_sym += '{:4d}'.format(i + 1) + '  '.join(['{:7.3f}'.format(s) for s in line])
        txt_sym += '\n'

    txt_sym += '\nBeta MOs: Symmetry Overlap Expectation Values\n'
    txt_sym += sep_line
    txt_sym += '     ' + '  '.join(['{:^7}'.format(s) for s in ir_labels])
    txt_sym += '\n'
    txt_sym += sep_line
    for i, line in enumerate(mo_soevs_b):
        txt_sym += '{:4d}'.format(i + 1) + '  '.join(['{:7.3f}'.format(s) for s in line])
        txt_sym += '\n'

    txt_sym += '\nWaveFunction: Symmetry Overlap Expectation Values\n'
    txt_sym += sep_line
    txt_sym += '     ' + '  '.join(['{:^7}'.format(s) for s in ir_labels])
    txt_sym += '\n'
    txt_sym += sep_line
    txt_sym += 'a-wf' + '  '.join(['{:7.3f}'.format(s) for s in wf_soevs_a])
    txt_sym += '\n'
    txt_sym += 'b-wf' + '  '.join(['{:7.3f}'.format(s) for s in wf_soevs_b])
    txt_sym += '\n'
    txt_sym += 'WFN ' + '  '.join(['{:7.3f}'.format(s) for s in wf_soevs])
    txt_sym += '\n'

    txt_sym += '\nWaveFunction: CSM-like values\n'
    txt_sym += sep_line
    txt_sym += '     ' + '  '.join(['{:^7}'.format(s) for s in ir_labels])
    txt_sym += '\n'
    txt_sym += sep_line

    txt_sym += 'Grim' + '  '.join(['{:7.3f}'.format(s) for s in grim_coef])
    txt_sym += '\n'
    txt_sym += 'CSM ' + '  '.join(['{:7.3f}'.format(s) for s in csm_coef])
    txt_sym += '\n'

    return txt_sym


def build_symmetry_ireducible_representation_analysis(parsed_data):

    ir_labels = parsed_data.IRLab
    mo_ir_alpha = parsed_data.mo_IRd_a
    mo_ir_beta = parsed_data.mo_IRd_b
    wf_ir_a = parsed_data.wf_IRd_a
    wf_ir_b = parsed_data.wf_IRd_b
    wf_ir = parsed_data.wf_IRd

    sep_line = '   ---------------------------------------------\n'

    txt_sym = '\nAlpha MOs: Irred. Rep. Decomposition\n'
    txt_sym += sep_line
    txt_sym += '     ' + '  '.join(['{:^7}'.format(s) for s in ir_labels])
    txt_sym += '\n'
    txt_sym += sep_line
    for i, line in enumerate(mo_ir_alpha):
        txt_sym += '{:4d}'.format(i + 1) + '  '.join(['{:7.3f}'.format(s) for s in line])
        txt_sym += '\n'

    txt_sym += '\nBeta MOs: Irred. Rep. Decomposition\n'
    txt_sym += sep_line
    txt_sym += '     ' + '  '.join(['{:^7}'.format(s) for s in ir_labels])
    txt_sym += '\n'
    txt_sym += sep_line
    for i, line in enumerate(mo_ir_beta):
        txt_sym += '{:4d}'.format(i + 1) + '  '.join(['{:7.3f}'.format(s) for s in line])
        txt_sym += '\n'

    txt_sym += '\nWaveFunction: Irred. Rep. Decomposition\n'
    txt_sym += sep_line
    txt_sym += '     ' + '  '.join(['{:^7}'.format(s) for s in ir_labels])
    txt_sym += '\n'
    txt_sym += sep_line
    txt_sym += 'a-wf' + '  '.join(['{:7.3f}'.format(s) for s in wf_ir_a])
    txt_sym += '\n'
    txt_sym += 'b-wf' + '  '.join(['{:7.3f}'.format(s) for s in wf_ir_b])
    txt_sym += '\n'
    txt_sym += 'WFN ' + '  '.join(['{:7.3f}'.format(s) for s in wf_ir])
    txt_sym += '\n'

    return txt_sym



from cosymlib.molecule import Molecule, Geometry
from cosymlib import file_io
from cosymlib.file_io import shape2file
from cosymlib.utils import get_shape_map, molecular_orbital_diagram, symmetry_energy_evolution
import sys


class Cosymlib:
    """
    Main class of cosymlib program that can perform all the jobs
    """

    def __init__(self, structures):

        self._molecules = []
        if isinstance(structures, list):
            for structure in structures:
                if isinstance(structure, Molecule):
                    self._molecules.append(structure)
                elif isinstance(structure, Geometry):
                    self._molecules.append(Molecule(structure))
                else:
                    raise AttributeError('Molecule object not found')
        else:
            if isinstance(structures, Molecule):
                self._molecules.append(structures)
            elif isinstance(structures, Geometry):
                self._molecules.append(Molecule(structures))
            else:
                raise AttributeError('Molecule object not found')

    def write_shape_measure_2file(self, shape_reference, central_atom=0, fix_permutation=False, output_name=None):
        """
        Method that prints to file shape's measure
        :param shape_reference: reference polyhedra label which user will compare with his polyhedra.
                            Reference labels can be found in [#f1]_
        :param central_atom: position of the central atom in molecule if exist
        :param output_name: custom name without extension
        :return: shape's measure in the output_name.tab file
        """

        if output_name is not None:
            output = open(output_name + '_tab.csv', 'w')
        else:
            output = sys.stdout

        molecules_name = [molecule.get_name() for molecule in self._molecules]
        shape_results_measures = []
        references = []
        for reference in shape_reference:
            if type(reference) is Geometry:
                shape_results_measures.append(self.get_shape_measure(reference.get_positions(), 'measure', central_atom,
                                                                     fix_permutation))
                references.append(reference.get_name())
            else:
                shape_results_measures.append(self.get_shape_measure(reference, 'measure', central_atom,
                                                                     fix_permutation))
                references.append(reference)

        output.write(file_io.header())
        output.write(shape2file.write_shape_measure_data(shape_results_measures, molecules_name, references))

    def write_shape_structure_2file(self, shape_reference, central_atom=0, fix_permutation=False, output_name=None):
        """
        Method that prints to file shape's structure
        :param shape_reference: reference polyhedra label which user will compare with his polyhedra.
                            Reference labels can be found in [#f1]_
        :param central_atom: position of the central atom in molecule if exist
        :param output_name: custom name without extension
        :return: shape's structure in the output_name.out file
        """

        if output_name is not None:
            output = open(output_name + '_shp.xyz', 'w')
        else:
            output = sys.stdout

        shape_results_structures = []
        references = []
        for reference in shape_reference:
            if type(reference) is Geometry:
                shape_results_structures.append(self.get_shape_measure(reference.get_positions(),
                                                                       'structure', central_atom, fix_permutation))
                references.append(reference.get_name())
            else:
                shape_results_structures.append(self.get_shape_measure(reference, 'structure', central_atom,
                                                                       fix_permutation))
                references.append(reference)

        self.write_shape_measure_2file(shape_reference, central_atom, fix_permutation, output_name)
        geometries = []
        for idm, molecule in enumerate(self._molecules):
            geometries.append(molecule.geometry)
            for idl, reference in enumerate(references):
                geometries.append(Geometry(symbols=molecule.geometry.get_symbols(),
                                           positions=shape_results_structures[idl][idm],
                                           name=molecule.get_name() + '_' + reference))
        output.write(file_io.write_file_xyz(geometries))

    def write_path_parameters_2file(self, shape_label1, shape_label2, central_atom=0,
                                    maxdev=15, mindev=0, maxgco=101, mingco=0, output_name=None):

        if output_name is not None:
            output = open(output_name + '_tab.csv', 'w')
        else:
            output = sys.stdout

        output.write(file_io.header())
        csm, devpath, GenCoord = self.get_path_parameters(shape_label1, shape_label2, central_atom=central_atom,
                                                          maxdev=maxdev, mindev=mindev, maxgco=maxgco, mingco=mingco)
        names_order = [molecule.get_name() for molecule in self._molecules]
        txt = shape2file.write_minimal_distortion_path_analysis(csm, devpath, GenCoord, maxdev, mindev,
                                                                mingco, maxgco, names_order)
        output.write(txt)

    def write_symgroup_measure_all_info(self, group, multi=1, central_atom=0, symbols=True, output_name=None):

        if output_name is not None:
            output = open(output_name + '.zout', 'w')
        else:
            output = sys.stdout

        results = self.get_symgroup_measure(group=group, multi=multi, symbols=symbols, central_atom=central_atom)
        txt = file_io.header()
        txt += file_io.symgroup_file.build_symgroup_data(group, [molecule.geometry for molecule in self._molecules],
                                                         results)
        output.write(txt)

    def write_symgroup_measure(self, group, multi=1, central_atom=0, symbols=True, output_name=None):

        if output_name is not None:
            output = open(output_name + '.ztab', 'w')
        else:
            output = sys.stdout

        results = self.get_symgroup_measure(group=group, multi=multi, symbols=symbols, central_atom=central_atom)
        txt = file_io.header()
        txt += file_io.symgroup_file.build_symgroup_measure(group, [molecule.geometry for molecule in self._molecules],
                                                            results)
        output.write(txt)

    def write_symgroup_structure(self, group, multi=1, central_atom=0, symbols=True, output_name=None):

        if output_name is not None:
            output = open(output_name + '_sym.xyz', 'w')
        else:
            output = sys.stdout

        self.write_symgroup_measure(group=group, multi=multi, symbols=symbols, central_atom=central_atom)
        results = self.get_symgroup_measure(group=group, multi=multi, symbols=symbols, central_atom=central_atom)
        geometries = []
        for idm, molecule in enumerate(self._molecules):
            geometries.append(Geometry(symbols=molecule.geometry.get_symbols(),
                                       positions=results[idm].nearest_structure,
                                       name=molecule.get_name() + '_' + group + ' with orientation ' +
                                            ' '.join('{:.8f}'.format(e) for e in results[idm].optimum_axis)))
        output.write(file_io.write_file_xyz(geometries))

    def write_wnfsym_measure_2file(self, group, vector_axis1=None, vector_axis2=None, center=None, output_name=None,
                                   n_molecule=0):
        if output_name is not None:
            output = open(output_name + '.wout', 'w')
        else:
            output = sys.stdout

        wfnsym_results = self.get_wfnsym_measure(group, vector_axis1, vector_axis2, center)
        txt = file_io.header()
        txt += file_io.wfnsym_file.build_symmetry_operated_matrices(group, self._molecules[n_molecule],
                                                                    wfnsym_results[n_molecule])
        txt += file_io.wfnsym_file.build_symmetry_overlap_analysis(wfnsym_results[n_molecule])
        txt += file_io.wfnsym_file.build_symmetry_ireducible_representation_analysis(wfnsym_results[n_molecule])
        output.write(txt)

    def write_wnfsym_sym_matrices_2file(self, group, vector_axis1=None, vector_axis2=None, center=None,
                                        output_name=None,
                                        n_molecule=0):
        if output_name is not None:
            output = open(output_name + '.wout', 'w')
        else:
            output = sys.stdout

        wfnsym_results = self.get_wfnsym_measure(group, vector_axis1, vector_axis2, center)
        txt = file_io.header()
        txt += file_io.wfnsym_file.build_symmetry_operated_matrices(group, self._molecules[0],
                                                                    wfnsym_results[n_molecule])
        output.write(txt)

    def write_wnfsym_sym_ovelap_2file(self, group, vector_axis1=None, vector_axis2=None, center=None, output_name=None,
                                      n_molecule=0):
        if output_name is not None:
            output = open(output_name + '.wout', 'w')
        else:
            output = sys.stdout

        wfnsym_results = self.get_wfnsym_measure(group, vector_axis1, vector_axis2, center)
        txt = file_io.header()
        txt += file_io.wfnsym_file.build_symmetry_overlap_analysis(wfnsym_results[n_molecule])
        output.write(txt)

    def write_wnfsym_ireducible_repr_2file(self, group, vector_axis1=None, vector_axis2=None, center=None,
                                           output_name=None, n_molecule=0):
        if output_name is not None:
            output = open(output_name + '.wout', 'w')
        else:
            output = sys.stdout

        wfnsym_results = self.get_wfnsym_measure(group, vector_axis1, vector_axis2, center)
        txt = file_io.header()
        txt += file_io.wfnsym_file.build_symmetry_ireducible_representation_analysis(wfnsym_results[n_molecule])
        output.write(txt)

    def write_mo_diagram(self, group, vector_axis1=None, vector_axis2=None, center=None, n_molecule=0):
        wfnsym_results = self.get_wfnsym_measure(group, vector_axis1, vector_axis2, center)
        molecular_orbital_diagram(self._molecules[0], wfnsym_results[n_molecule])

    def write_sym_energy_evolution(self, group, vector_axis1=None, vector_axis2=None, center=None):
        wfnsym_results = self.get_wfnsym_measure(group, vector_axis1, vector_axis2, center)
        symmetry_energy_evolution(self._molecules, wfnsym_results)

    def get_shape_measure(self, label, kind, central_atom=0, fix_permutation=False):
        get_measure = 'get_shape_' + kind
        return [getattr(molecule.geometry, get_measure)(label, central_atom=central_atom,
                                                        fix_permutation=fix_permutation)
                for molecule in self._molecules]

    def get_molecule_path_deviation(self, shape_label1, shape_label2, central_atom=0):
        return [molecule.geometry.get_path_deviation(shape_label1, shape_label2, central_atom) for molecule
                in self._molecules]

    def get_molecule_generalized_coord(self, shape_label1, shape_label2, central_atom=0):
        return [molecule.geometry.get_generalized_coordinate(shape_label1, shape_label2, central_atom)
                for molecule in self._molecules]

    def get_path_parameters(self, shape_label1, shape_label2, central_atom=0, maxdev=15, mindev=0,
                            maxgco=101, mingco=0):

        def filter_results(results, criteria, max, min):
            filter = [True if x <= max and x >= min else False for x in criteria]
            results = [i for indx, i in enumerate(results) if filter[indx] == True]
            return results

        if type(shape_label1) is Geometry:
            label1 = shape_label1.get_positions()
            label1_name = shape_label1.get_name()
        else:
            label1 = shape_label1
            label1_name = shape_label1
        if type(shape_label2) is Geometry:
            label2 = shape_label2.get_positions()
            label2_name = shape_label2.get_name()
        else:
            label2 = shape_label2
            label2_name = shape_label2

        csm = {label1_name: self.get_shape_measure(label1, 'measure', central_atom),
               label2_name: self.get_shape_measure(label2, 'measure', central_atom)}
        devpath = self.get_molecule_path_deviation(label1, label2, central_atom)
        generalized_coord = self.get_molecule_generalized_coord(label1, label2, central_atom)
        criteria = devpath
        devpath = filter_results(devpath, criteria, maxdev, mindev)
        generalized_coord = filter_results(generalized_coord, criteria, maxdev, mindev)
        criteria = generalized_coord
        devpath = filter_results(devpath, criteria, maxgco, mingco)
        generalized_coord = filter_results(generalized_coord, criteria, maxgco, mingco)
        return csm, devpath, generalized_coord

    def get_symgroup_measure(self, group, multi=1, central_atom=0, symbols=True):
        return [molecule.geometry.get_symmetry_measure(label=group, multi=multi, central_atom=central_atom,
                                                       symbols=symbols) for molecule in self._molecules]

    def get_wfnsym_measure(self, group, vector_axis1, vector_axis2, center):
        return [molecule.get_mo_symmetry(group, vector_axis1=vector_axis1, vector_axis2=vector_axis2, center=center)
                for molecule in self._molecules]

    def write_minimum_distortion_path_shape_2file(self, shape_label1, shape_label2, central_atom=0,
                                                  num_points=20, output_name=None):

        self.write_path_parameters_2file(shape_label1, shape_label2, central_atom=central_atom, output_name=output_name)
        if output_name is not None:
            output = open(output_name + '_pth.csv', 'w')
            output2 = open(output_name + '_pth.xyz', 'w')
        else:
            output = sys.stdout
            output2 = sys.stdout

        if type(shape_label1) is Geometry:
            label1 = shape_label1.get_positions()
            label1_name = shape_label1.get_name()
        else:
            label1 = shape_label1
            label1_name = shape_label1
        if type(shape_label2) is Geometry:
            label2 = shape_label2.get_positions()
            label2_name = shape_label2.get_name()
        else:
            label2 = shape_label2
            label2_name = shape_label2

        path_parameters = self.get_path_parameters(shape_label1, shape_label2, central_atom=central_atom)[0]
        path = get_shape_map(label1, label2, num_points)
        output.write(shape2file.write_shape_map(label1_name, label2_name, path))
        test_structures = []
        for ids, structure in enumerate(path[2]):
            test_structures.append(Geometry(symbols=['' for _ in range(len(structure))],
                                            positions=structure, name='map_structure{}'.format(ids)))
        output2.write(file_io.write_file_xyz(test_structures))
        if output_name is None:
            import matplotlib.pyplot as plt
            plt.plot(path[0], path[1], 'k', linewidth=2.0)
            plt.scatter(path_parameters[label1_name], path_parameters[label2_name], linewidths=0.01)
            plt.xlabel(label1_name)
            plt.ylabel(label2_name)
            plt.show()

    def write_point_group(self, tol=0.01):
        return [molecule.get_pointgroup(tol) for molecule in self._molecules]



import sys
from cosymlib import tools


def build_symgroup_data(label, geometries, symgroup_results):
    # if output_name is not None:
    #     output = open(output_name + '.zout', 'w')
    #     output2 = open(output_name + '.ztab', 'w')
    # else:
    #     output = sys.stdout
    #     output2 = sys.stdout

    sep_line = '..................................................\n'

    sym_txt = 'Evaluating symmetry operation : {}\n'.format(label)

    for idx, geometry in enumerate(geometries):
        sym_txt += '{}\n'.format(geometry.get_name())
        sym_txt += '\n'
        sym_txt += 'Centered Structure\n'
        sym_txt += sep_line
        center_mass = tools.center_mass(geometry.get_symbols(), geometry.get_positions())
        for idn, array in enumerate(geometry.get_positions()):
            array = array - center_mass
            sym_txt += '{:2} {:12.8f} {:12.8f} {:12.8f}\n'.format(geometry.get_symbols()[idn],
                                                                  array[0], array[1], array[2])
        sym_txt += sep_line

        sym_txt += 'Optimal permutation\n'
        for idn, permutation in enumerate(symgroup_results[idx].optimum_permutation):
            sym_txt += '{:2} {:2}\n'.format(idn + 1, permutation)
        sym_txt += '\n'

        sym_txt += 'Inverted structure\n'
        for idn, axis in enumerate(symgroup_results[idx].nearest_structure):
            sym_txt += '{:2} {:12.8f} {:12.8f} {:12.8f}\n'.format(geometry.get_symbols()[idn],
                                                                    axis[0], axis[1], axis[2])
        sym_txt += '\n'

        sym_txt += 'Reference axis\n'
        for array in symgroup_results[idx].reference_axis:
            sym_txt += '{:12.8f} {:12.8f} {:12.8f}\n'.format(array[0], array[1], array[2])
        sym_txt += '\n'

        sym_txt += 'Symmetry measure {:.5f}\n'.format(symgroup_results[idx].csm)
        sym_txt += sep_line

    return sym_txt


def build_symgroup_measure(label, geometries, symgroup_results):

    sym_txt = 'Evaluating symmetry operation : {}\n \n'.format(label)
    for idx, geometry in enumerate(geometries):
        csm = symgroup_results[idx].csm
        max_name = len(max(geometry.get_name(), key=len))
        sym_txt += '{}'.format(geometry.get_name())
        if max_name < 9:
            n = 18 - len(geometry.get_name())
        else:
            n = 9 + max_name - len(geometry.get_name())
        sym_txt += '{:{width}.{prec}f}\n'.format(csm, width=n, prec=3)

    return sym_txt


import numpy as np
from cosymlib.tools import element_mass


def get_mass(symbols):
    mass_vector = []
    for symbol in symbols:
        mass_vector.append(element_mass(symbol))
    return mass_vector


def get_center_mass(symbols, coordinates):
    mass_vector = get_mass(symbols)
    cbye = [np.dot(mass_vector[i], coordinates[i]) for i in range(len(symbols))]
    r = np.sum(cbye, axis=0)
    r = r / np.sum(mass_vector)
    return r


def get_inertia_tensor(elements, coord):
    mass_vector = get_mass(elements)
    total_inertia = 0

    inertia_xx = np.dot(mass_vector, (np.square(coord[:, 1]) + np.square(coord[:, 2])))
    inertia_yy = np.dot(mass_vector, (np.square(coord[:, 0]) + np.square(coord[:, 2])))
    inertia_zz = np.dot(mass_vector, (np.square(coord[:, 0]) + np.square(coord[:, 1])))
    inertia_xy = -(np.dot(mass_vector, np.multiply(coord[:, 0], coord[:, 1])))
    inertia_yz = -(np.dot(mass_vector, np.multiply(coord[:, 1], coord[:, 2])))
    inertia_xz = -(np.dot(mass_vector, np.multiply(coord[:, 0], coord[:, 2])))

    inertia_matrix = ([[inertia_xx, inertia_xy, inertia_xz],
                       [inertia_xy, inertia_yy, inertia_yz],
                       [inertia_xz, inertia_yz, inertia_zz]])

    for idx, atom in enumerate(coord):
        total_inertia += mass_vector[idx] * np.dot(atom, atom)

    inertia_matrix /= total_inertia
    return inertia_matrix




def write_shape_measure_data(measures, molecules_name, shape_label):

    txt_shape = '{}'.format('Structure')
    max_name = len(max(molecules_name, key=len))
    if max_name < 9:
        n = 5
    else:
        n = max_name - 4
    for label in shape_label:
        n += len(label)
        txt_shape += '{}'.format(label.rjust(n))
        n = 12 - len(label)
    txt_shape += '\n\n'

    for idx, name in enumerate(molecules_name):
        max_name = len(max(molecules_name, key=len))
        txt_shape += '{}'.format(name)
        if max_name < 9:
            n = 18 - len(name)
        else:
            n = 9 + max_name - len(name)
        for idn, label in enumerate(shape_label):
            txt_shape += ', {:{width}.{prec}f}'.format(measures[idn][idx], width=n, prec=3)
            n = 11
        txt_shape += '\n'
    txt_shape += '\n'

    return txt_shape


def write_shape_map(shape_label1, shape_label2, path):

    txt_shape = " {:6} {:6}\n".format(shape_label1, shape_label2)
    for idx, value in enumerate(path[0]):
        txt_shape += '{:6.3f}, {:6.3f}'.format(path[0][idx], path[1][idx])
        txt_shape += '\n'
    return txt_shape


def write_minimal_distortion_path_analysis(measures, pathdev, GenCoord, maxdev, mindev, mingco, maxgco, molecules_name):

    txt_shape = "Deviation threshold to calculate Path deviation function: {:2.1f}% - {:2.1f}%\n".format(mindev, maxdev)
    txt_shape += "Deviation threshold to calculate Generalized Coordinate: {:2.1f}% - {:2.1f}%\n".format(mingco, maxgco)
    txt_shape += "\n"
    txt_shape += '{:}'.format('structure'.upper())
    txt_shape += " {:>7} {:>9}".format(list(measures.keys())[0], list(measures.keys())[1])
    txt_shape += "{:>12} {:>9}".format('DevPath', 'GenCoord')
    txt_shape += "\n"

    for idx, molecule_name in enumerate(molecules_name):
        txt_shape += '{}  ,'.format(molecule_name)
        if molecule_name.strip() == '':
            width = 6 + len(molecule_name)
        else:
            width = 14 - len(molecule_name)
        for label in list(measures.keys()):
            txt_shape += ' {:{width}.{prec}f},'.format(measures[label][idx], width=width, prec=3)
            width = 7
        txt_shape += '{:8.1f}, {:8}'.format(pathdev[idx], GenCoord[idx])
        txt_shape += '\n'
    return txt_shape




import sys
from cosymlib.molecule import Geometry
from cosymlib.file_io import read_input_file
import os.path


def read_old_input(file_name):
    """
    Reads the old Shape's program input
    :param file_name: file name
    :return: list of Geometry objects and options
    """
    input_molecule = [[], []]
    structures = []
    options = {'%out': None, '%conquest': None, '%external': False, '%fullout': False, '%test': False,
               '%n_atoms': 0, '%central_atom': 0, '%labels': 0, '%path': False}

    with open(file_name, mode='r') as lines:
        while True:
            line = lines.readline().split()
            if '$' in line or '!' in line:
                pass
            elif any('%' in word for word in line):
                if len(line) > 1:
                    options[line[0]] = line[1]
                else:

                    options[line[0]] = True
            else:
                try:
                    int(line[0])
                    if options['%n_atoms'] == 0:
                        options['%n_atoms'] = int(line[0])
                        options['%central_atom'] = int(line[1])
                    else:
                        options['%labels'] = line
                except (ValueError, IndexError):
                    break

        n_atoms = options['%n_atoms']
        if options['%central_atom'] != 0:
            n_atoms += 1
        if options['%conquest'] is not None:
            dir = os.path.dirname(os.path.abspath(file_name))
            structures = read_input_file(os.path.join(dir, options['%conquest'] + '.cor'))
        else:
            while True:
                if not line:
                    break
                name = line[0]
                for i in range(n_atoms):
                    line = lines.readline().split()
                    if '!' in line:
                        pass
                    else:
                        if len(line) == 4:
                            input_molecule[0].append(line[0])
                            input_molecule[1].append(line[1:])
                        elif len(line) == 5:
                            input_molecule[0].append(line[0])
                            input_molecule[1].append(line[1:-1])
                        else:
                            sys.exit('Wrong input format')
                if input_molecule[0]:
                    structures.append(Geometry(symbols=input_molecule[0],
                                               positions=input_molecule[1],
                                               name=name))
                    input_molecule = [[], []]
                line = lines.readline().split()

    return [structures, options]



from wfnsympy import WfnSympy
from cosymlib import tools
import hashlib


class Wfnsym:

    def __init__(self, molecule):

        self._molecule = molecule
        self._Ne_val = self._get_valence_electrons()
        self._wfnsym_dict = self._molecule.electronic_structure.basis
        self._results = {}

    def symmetry_overlap_analysis(self,
                                  group,
                                  vector_axis1,
                                  vector_axis2,
                                  center):

        hash = hashlib.md5('{}{}'.format(group, vector_axis1, vector_axis2, center).encode()).hexdigest()
        if hash not in self._results:
            self._do_measure(group, vector_axis1, vector_axis2, center)
        return [self._results[hash].ideal_gt, self._results[hash].SymLab, self._results[hash].mo_SOEVs_a,
                self._results[hash].mo_SOEVs_b, self._results[hash].wf_SOEVs_a, self._results[hash].wf_SOEVs_b,
                self._results[hash].wf_SOEVs, self._results[hash].grim_coef, self._results[hash].csm_coef]

    def symmetry_ireducible_representation_analysis(self, group, vector_axis1, vector_axis2, center):

        hash = hashlib.md5('{}{}'.format(group, vector_axis1, vector_axis2, center).encode()).hexdigest()
        if hash not in self._results:
            self._do_measure(group, vector_axis1, vector_axis2, center)
        return [self._results[hash].IRLab, self._results[hash].mo_IRd_a, self._results[hash].mo_IRd_b,
                self._results[hash].wf_IRd_a, self._results[hash].wf_IRd_b, self._results[hash].wf_IRd]

    def symmetry_matrix(self, group, vector_axis1, vector_axis2, center):

        hash = hashlib.md5('{}{}'.format(group, vector_axis1, vector_axis2, center).encode()).hexdigest()
        if hash not in self._results:
            self._do_measure(group, vector_axis1, vector_axis2, center)
        return self._results[hash].SymMat

    def _do_measure(self, group, vector_axis1, vector_axis2, center):

        hash = hashlib.md5('{}{}'.format(group, vector_axis1, vector_axis2, center).encode()).hexdigest()
        self._results[hash] = WfnSympy(coordinates=self._molecule.geometry.get_positions(),
                                       symbols=self._molecule.geometry.get_symbols(),
                                       basis=self._wfnsym_dict,
                                       center=center, VAxis=vector_axis1, VAxis2=vector_axis2,
                                       alpha_mo_coeff=self._molecule.electronic_structure.coefficients_a,
                                       beta_mo_coeff=self._molecule.electronic_structure.coefficients_b,
                                       charge=self._molecule.electronic_structure.charge,
                                       multiplicity=self._molecule.electronic_structure.multiplicity,
                                       group=group.upper(),
                                       valence_only=self._molecule.electronic_structure.valence_only)

    def results(self, group, vector_axis1, vector_axis2, center):

        hash = hashlib.md5('{}{}'.format(group, vector_axis1, vector_axis2, center).encode()).hexdigest()
        if hash not in self._results:
            self._do_measure(group, vector_axis1, vector_axis2, center)
        return self._results[hash]

    def _get_valence_electrons(self):
        n_valence = 0
        for symbol in self._molecule.geometry.get_symbols():
            n_valence += tools.element_valence_electron(symbol)
        return n_valence



import os
import sys
import re
from cosymlib.molecule import Molecule, Geometry, ElectronicStructure
import numpy as np
from cosymlib import tools
from cosymlib.file_io import shape2file, wfnsym_file, symgroup_file
from cosymlib import __version__


def nonblank_lines(f):
    for l in f:
        line = l.rstrip()
        if line:
            yield line


# INPUT part
def read_input_file(input_name):
    # print('Reading file {}...'.format(os.path.basename(input_name)))
    if os.stat(input_name).st_size == 0:
        raise FileExistsError('File {} is empty'.format(os.path.basename(input_name)))
    file_name, file_extension = os.path.splitext(input_name)
    if 'molden' in file_name:
        file_extension = ' molden'
    method_name = 'get_molecule_from_file_' + file_extension[1:]
    possibles = globals().copy()
    possibles.update(locals())
    method = possibles.get(method_name)
    if not method:
        raise NotImplementedError("Method {} is not implemented".format(method_name))

    return method(input_name)


def get_molecule_from_file_xyz(file_name):
    """
    Reads a XYZ file and returns the geometry of all structures in it
    :param file_name: file name
    :return: list of Geometry objects
    """
    input_molecule = [[], []]
    molecules = []
    with open(file_name, mode='r') as lines:
        lines.readline()
        name = lines.readline()
        if name.strip():
            name = name.split()[0]
        for line in lines:
            if '$' in line or '#' in line:
                pass
            else:
                try:
                    float(line.split()[1])
                    input_molecule[0].append(line.split()[0])
                    input_molecule[1].append(line.split()[1:])
                except (ValueError, IndexError):
                    if input_molecule[0]:
                        molecules.append(Geometry(symbols=input_molecule[0],
                                                  positions=input_molecule[1],
                                                  name=name))
                    input_molecule = [[], []]
                    name = line.split()[0]

        molecules.append(Geometry(symbols=input_molecule[0],
                                  positions=input_molecule[1],
                                  name=name))

    return molecules


def get_molecule_from_file_cor(file_name):
    """
    Reads a Conquest formatted file and the geometry of all structures in it
    :param file_name: file name
    :return: list of Geometry objects
    """
    input_molecule = [[], []]
    structures = []
    with open(file_name, mode='r') as lines:
        line = lines.readline()
        name = line.replace('**FRAG**', '')[:-1]
        name = name.replace(' ', '')
        for line in lines:
            if '$' in line or '#' in line:
                pass
            else:
                try:
                    float(line.split()[1])
                    if len(line.split()) == 4:
                        input_molecule[0].append(line.split()[0])
                        input_molecule[1].append(line.split()[1:])
                    elif len(line.split()) == 5:
                        input_molecule[0].append(line.split()[0])
                        input_molecule[1].append(line.split()[1:-1])
                    else:
                        float(line.split()[2])
                except (ValueError, IndexError):
                    if input_molecule:
                        structures.append(Geometry(symbols=input_molecule[0],
                                                   positions=input_molecule[1],
                                                   name=name))
                    input_molecule = [[], []]
                    name = line.replace('**FRAG**', '')[:-1]
                    name = name.replace(' ', '')
        structures.append(Geometry(symbols=input_molecule[0],
                                   positions=input_molecule[1],
                                   name=name))
    return structures


def get_molecule_from_file_fchk(file_name):
    key_list = ['Charge', 'Multiplicity', 'Atomic numbers', 'Current cartesian coordinates',
                'Shell type', 'Number of primitives per shell', 'Shell to atom map', 'Primitive exponents',
                'Contraction coefficients', 'P(S=P) Contraction coefficients',
                'Alpha MO coefficients', 'Beta MO coefficients']
    input_molecule = [[] for _ in range(len(key_list))]
    read = False
    with open(file_name, mode='r') as lines:
        name = lines.readline()
        line = lines.readline().split()
        basis_set = line[-1]
        if 'R' in line[1]:
            del key_list[-1]

        n = 1
        options = True
        for line in lines:
            if read:
                try:
                    float(line.split()[0])
                    input_molecule[n].append(line.split())
                except ValueError:
                    read = False

            for idn, key in enumerate(key_list):
                if key in line:
                    if n == len(key_list) - 1:
                        break
                    if options and idn != 2:
                        input_molecule[idn].append(int(line.split()[-1]))
                        n = idn
                        break
                    else:
                        options = False
                    if n == idn:
                        n += 1
                    else:
                        n = idn
                    read = True
                    break

        for n in range(2, len(input_molecule) - 1):
            input_molecule[n] = reformat_input(input_molecule[n])
        bohr_to_angstrom = 0.529177249
        coordinates = np.array(input_molecule[3], dtype=float).reshape(-1, 3) * bohr_to_angstrom

        atomic_number = [int(num) for num in input_molecule[2]]
        symbols = []
        for number in atomic_number:
            symbols.append(tools.atomic_number_to_element(number))

        basis = basis_format(basis_set_name=basis_set,
                             atomic_numbers=atomic_number,
                             atomic_symbols=symbols,
                             shell_type=input_molecule[4],
                             n_primitives=input_molecule[5],
                             atom_map=input_molecule[6],
                             p_exponents=input_molecule[7],
                             c_coefficients=input_molecule[8],
                             p_c_coefficients=input_molecule[9])

        geometry = Geometry(symbols=symbols,
                            positions=coordinates,
                            name=name)

        Ca = np.array(input_molecule[10], dtype=float).reshape(-1, int(np.sqrt(len(input_molecule[10]))))
        if input_molecule[11]:
            Cb = np.array(input_molecule[11], dtype=float).reshape(-1, int(np.sqrt(len(input_molecule[11]))))
        else:
            Cb = []
        ee = ElectronicStructure(charge=input_molecule[0][0],
                                 multiplicity=input_molecule[1][0],
                                 basis=basis,
                                 orbital_coefficients=[Ca, Cb])

        return [Molecule(geometry, ee)]


def get_molecule_from_file_molden(file_name):
    key_list = ['Charge', 'Multiplicity', 'Atomic numbers', 'Current cartesian coordinates',
                'Shell type', 'Number of primitives per shell', 'Shell to atom map', 'Primitive exponents',
                'Contraction coefficients', 'P(S=P) Contraction coefficients',
                'Alpha MO coefficients', 'Beta MO coefficients', 'MO Energies']

    type_list = {'s': '0',
                 'p': '1',
                 'd': '2',
                 'f': '3',
                 'sp': '-1'}

    input_molecule = {key: [] for key in key_list}
    read_molden = False
    read_coordinates = False
    read_basis = False
    read_coefficients = False
    occupation = {'Alpha': [], 'Beta': []}
    with open(file_name, mode='r') as lines:
        lines.readline()
        lines.readline()
        name = lines.readline()
        for line in nonblank_lines(lines):

            if '[' in line:
                pass
            elif read_molden:
                if '======= END OF MOLDEN-FORMAT MOLECULAR ORBITALS =======' in line:
                    break
                if read_coordinates:
                    input_molecule['Atomic numbers'].append(line.split()[2])
                    input_molecule['Current cartesian coordinates'].append(line.split()[3:])
                if read_basis:
                    try:
                        number = float(line.split()[0].replace('D', 'E'))
                        if number-int(number) != 0.:
                            input_molecule['Primitive exponents'].append(line.split()[0].replace('D', 'E'))
                            input_molecule['Contraction coefficients'].append(line.split()[1].replace('D', 'E'))
                            if len(line.split()) > 2:
                                input_molecule['P(S=P) Contraction coefficients'].append(line.split()[2].replace('D', 'E'))
                            else:
                                input_molecule['P(S=P) Contraction coefficients'].append(0.)
                        else:
                            atom = line.split()[0]
                    except ValueError:
                        input_molecule['Shell to atom map'].append(atom)
                        input_molecule['Shell type'].append(type_list[line.split()[0].lower()])
                        input_molecule['Number of primitives per shell'].append(line.split()[1])
                if read_coefficients:
                    if 'Sym' in line:
                        pass
                    else:
                        if 'Ene' in line:
                            input_molecule['MO Energies'].append(float(re.split('[= ]', line)[-1]))
                        elif 'Spin' in line:
                            spin = re.split('[= ]', line)[-1]
                            input_molecule[spin + ' MO coefficients'].append([])
                        elif 'Occup' in line:
                            occupation[spin].append(float(line.split('=')[-1]))
                        else:
                            input_molecule[spin+' MO coefficients'][-1].append(line.split()[1])

            if '[Atoms]' in line:
                read_molden = True
                read_coordinates = True
                if 'AU' in line:
                    bohr_to_angstrom = 0.529177249
                else:
                    bohr_to_angstrom = 1
            elif '[GTO]' in line:
                read_coordinates = False
                read_basis = True
            elif '[MO]' in line:
                read_basis = False
                read_coefficients = True
            elif '[5D]' in line:
                input_molecule['Shell type'] = [x.replace('2', '-2') for x in input_molecule['Shell type']]
            elif '[7F]' in line:
                input_molecule['Shell type'] = [x.replace('3', '-3') for x in input_molecule['Shell type']]

        coordinates = np.array(input_molecule['Current cartesian coordinates'], dtype=float).reshape(-1, 3) * bohr_to_angstrom

        total_n_electrons = 0
        symbols = []
        for atom_number in input_molecule['Atomic numbers']:
            total_n_electrons += int(atom_number)
            symbols.append(tools.atomic_number_to_element(atom_number))

        occupation['Alpha'] = np.array(occupation['Alpha'], dtype=float)
        occupation['Beta'] = np.array(occupation['Beta'], dtype=float)
        if len(occupation['Beta']) == 0:
            if 2.0 not in occupation['Alpha']:
                occupation['Alpha'] = 2*occupation['Alpha']
            n_electrons = sum(occupation['Alpha'])
            input_molecule['Multiplicity'].append(sum(map(lambda x: x % 2 == 1, occupation['Alpha'])) + 1)
        else:
            if 2.0 not in occupation['Beta']:
                occupation['Beta'] = 2*occupation['Beta']
            n_electrons = sum(occupation['Alpha'] + occupation['Beta'])
            input_molecule['Multiplicity'].append(sum(occupation['Alpha'] - occupation['Beta']) + 1)

        input_molecule['Charge'].append(int(total_n_electrons - n_electrons))

        basis = basis_format(basis_set_name='UNKNOWN',
                             atomic_numbers=input_molecule['Atomic numbers'],
                             atomic_symbols=symbols,
                             shell_type=input_molecule['Shell type'],
                             n_primitives=input_molecule['Number of primitives per shell'],
                             atom_map=input_molecule['Shell to atom map'],
                             p_exponents=input_molecule['Primitive exponents'],
                             c_coefficients=input_molecule['Contraction coefficients'],
                             p_c_coefficients=input_molecule['P(S=P) Contraction coefficients'])

        geometry = Geometry(symbols=symbols,
                            positions=coordinates,
                            name=name)

        Ca = np.array(input_molecule['Alpha MO coefficients'], dtype=float)
        if input_molecule['Beta MO coefficients']:
            Cb = np.array(input_molecule['Beta MO coefficients'], dtype=float)
        else:
            Cb = []

        ee = ElectronicStructure(charge=input_molecule['Charge'][0],
                                 multiplicity=input_molecule['Multiplicity'][0],
                                 basis=basis,
                                 orbital_coefficients=[Ca, Cb],
                                 mo_energies=input_molecule['MO Energies'])

        return [Molecule(geometry, ee)]


def get_molecule_from_file_ref(file_name):
    """
    Reads a Conquest formatted file and the geometry of all structures in it
    :param file_name: file name
    :return: list of Geometry objects
    """
    input_molecule = []
    structures = []
    with open(file_name, mode='r') as lines:
        lines.readline()
        lines.readline()
        name = lines.readline().split()[0]
        for line in lines:
            if '$' in line or '#' in line:
                pass
            else:
                try:
                    if len(line.split()) > 3:
                        float(line.split()[1])
                        input_molecule.append(line.split()[1:])
                    else:
                        float(line.split()[0])
                        input_molecule.append(line.split())
                except (ValueError, IndexError):
                    if input_molecule:
                        structures.append(Geometry(positions=input_molecule,
                                                  name=name))
                    input_molecule = []
                    name = line.split()[0]
        structures.append(Geometry(positions=input_molecule,
                                  name=name))
    return structures


def basis_format(basis_set_name,
                 atomic_numbers,
                 atomic_symbols,
                 shell_type,
                 n_primitives,
                 atom_map,
                 p_exponents,
                 c_coefficients,
                 p_c_coefficients):
    type_list = {'0': ['s', 1],
                 '1': ['p', 3],
                 '2': ['d', 6],
                 '3': ['f', 10],
                 '-1': ['sp', 4],
                 '-2': ['d_', 5],
                 '-3': ['f_', 7]}

    atomic_numbers = np.array(atomic_numbers, dtype=int)
    atom_map = np.array(atom_map, dtype=int)

    basis_set = {'name': basis_set_name,
                 'primitive_type': 'gaussian'}

    # shell_type_index = [0] + np.cumsum([type_list['{}'.format(s)][1]
    #                                     for s in shell_type]).tolist()
    prim_from_shell_index = [0] + np.cumsum(np.array(n_primitives, dtype=int)).tolist()

    atoms_data = []
    for iatom, atomic_number in enumerate(atomic_numbers):
        symbol = atomic_symbols[iatom]

        shell_from_atom_counts = np.unique(atom_map, return_counts=True)[1]
        shell_from_atom_index = np.unique(atom_map, return_index=True)[1]

        shells_data = []
        for ishell in range(shell_from_atom_counts[iatom]):
            st = type_list['{}'.format(shell_type[shell_from_atom_index[iatom] + ishell])]
            ini_prim = prim_from_shell_index[shell_from_atom_index[iatom] + ishell]
            fin_prim = prim_from_shell_index[shell_from_atom_index[iatom] + ishell + 1]

            shells_data.append({
                'shell_type': st[0],
                'p_exponents': p_exponents[ini_prim: fin_prim],
                'con_coefficients': c_coefficients[ini_prim: fin_prim],
                'p_con_coefficients': p_c_coefficients[ini_prim: fin_prim],
            })

        atoms_data.append({'shells': shells_data,
                           'symbol': symbol,
                           'atomic_number': atomic_number})

    basis_set['atoms'] = atoms_data

    return basis_set


# OUTPUT part
def header():
    txt_header = '-' * 70 + '\n'
    txt_header += ' COSYM v{}\n Electronic Structure Group,  Universitat de Barcelona\n'.format(__version__)
    txt_header += '-' * 70 + '\n' + '\n'
    return txt_header


def write_input_info(initial_geometries, output_name=None):
    if output_name is not None:
        output = open(output_name + '.tab', 'w')
    else:
        output = sys.stdout
    output.write(header())

    for ids, geometry in enumerate(initial_geometries):
        output.write('Structure {} : {}\n'.format(ids+1, geometry.get_name()))

        for idn, array in enumerate(geometry.get_positions()):
            output.write('{:2s}'.format(geometry.get_symbols()[idn]))
            output.write(' {:11.7f} {:11.7f} {:11.7f}\n'.format(array[0], array[1], array[2]))
        output.write('\n')


def write_file_xyz(geometries):

    txt = ''
    for geometry in geometries:
        txt += '{}\n'.format(geometry.get_n_atoms())
        txt += '{}\n'.format(geometry.get_name())
        for idp, position in enumerate(geometry.get_positions()):
            txt += '{:2} {:11.6f} {:11.6f} {:11.6f}\n'.format(geometry.get_symbols()[idp],
                                                              position[0], position[1], position[2])
    return txt


def reformat_input(array):
    flat_list = []
    for sublist in array:
        for item in sublist:
            if len(item) > 2:
                flat_list.append(item)
            else:
                flat_list.append(item)
    return flat_list



from cosymlib import shape, tools
from cosymlib.symmetry import symgroup
import numpy as np


class Geometry:
    def __init__(self,
                 symbols=[],
                 positions=None,
                 name=None):

        self._central_atom = None
        self._symbols = []
        self._positions = []
        self._atom_groups = list(symbols)

        if name.strip():
            self._name = name
        else:
            self._name = ' ' * 5

        for symbol in symbols:
            try:
                int(symbol)
                self._symbols.append(tools.atomic_number_to_element(int(symbol)))
            except (ValueError, TypeError):
                self._symbols.append(symbol.capitalize())
                for ida, a in enumerate(symbol):
                    try:
                        int(a)
                        self._symbols[-1] = self._symbols[-1][:ida]
                        break
                    except (ValueError, TypeError, IndexError):
                        pass

        try:
            float(positions[0])
            for symbol in positions:
                self._positions.append(float(symbol))
            self._positions = list(chunks(self._positions, 3))
        except (ValueError, TypeError, IndexError):
            for symbol in positions:
                self._positions.append([float(j) for j in symbol])

        self._positions = np.array(self._positions)
        self._shape = shape.Shape(self)
        self._symgroup = symgroup.Symgroup(self)

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    def get_positions(self):
        return self._positions

    def get_n_atoms(self):
        return len(self.get_positions())

    def get_symbols(self):
        return self._symbols

    def get_shape_measure(self, shape_label, central_atom=0, fix_permutation=False):
        return self._shape.measure(shape_label, central_atom=central_atom, fix_permutation=fix_permutation)

    def get_shape_structure(self, shape_label, central_atom=0, fix_permutation=False):
        return self._shape.structure(shape_label, central_atom=central_atom, fix_permutation=fix_permutation)

    def get_symmetry_measure(self, label, central_atom=0, multi=1, symbols=True):
        return self._symgroup.results(label, central_atom=central_atom, multi=multi)

    def get_path_deviation(self, shape_label1, shape_label2, central_atom=0):
        return self._shape.get_path_deviation(shape_label1, shape_label2, central_atom)

    def get_generalized_coordinate(self, shape_label1, shape_label2, central_atom=0):
        return self._shape.get_generalized_coordinate(shape_label1, shape_label2, central_atom)


def chunks(l, n):
    for i in range(0, len(l), n):
        yield l[i:i+n]



import huckelpy


class ExtendedHuckel:

    def __init__(self, geometry):
        self._EH = huckelpy.ExtendedHuckel(geometry.get_positions(), geometry.get_symbols())

    def get_mo_coefficients(self):
        return self._EH.get_eigenvectors()

    def get_basis(self):
        return self._EH.get_molecular_basis()

    def get_mo_energies(self):
        return self._EH.get_mo_energies()

    def build_fchk_file(self, name):
        txt_fchk = huckelpy.file_io.build_fchk(self._EH)
        open(name + '.fchk', 'w').write(txt_fchk)



class ElectronicStructure:
    def __init__(self,
                 charge=0,
                 multiplicity=1,
                 basis=None,
                 orbital_coefficients=None,
                 mo_energies=None,
                 valence_only=False):

        self._charge = charge
        self._multiplicity = multiplicity
        self._basis = basis
        self._valence_only = valence_only
        self._Ca = orbital_coefficients[0]
        if not orbital_coefficients[1]:
            self._Cb = None
        else:
            self._Cb = orbital_coefficients[1]
        self._mo_energies = mo_energies

    @property
    def charge(self):
        return self._charge

    @property
    def multiplicity(self):
        return self._multiplicity

    @property
    def basis(self):
        return self._basis

    @property
    def coefficients_a(self):
        return self._Ca

    @property
    def coefficients_b(self):
        return self._Cb

    @property
    def energies(self):
        return self._mo_energies

    @property
    def valence_only(self):
        return self._valence_only



from numpy.distutils.core import setup, Extension

def get_version_number():
    for l in open('cosymlib/__init__.py', 'r').readlines():
        if not(l.find('__version__')):
            exec(l, globals())
            return __version__


shape = Extension('cosymlib.shape.shp',
                  # extra_compile_args=['-std=c99'],
                  #include_dirs=include_dirs_numpy,
                  sources=['fortran/shp.pyf', 'fortran/shp.f90'])

setup(name='cosymlib',
      version=get_version_number(),
      description='Continuous measures of shape and symmetry',
      author='Efrem Bernuz',
      author_email='komuisan@gmail.com',
      packages=['cosymlib',
                'cosymlib.shape',
                'cosymlib.molecule',
                'cosymlib.molecule.geometry',
                'cosymlib.molecule.electronic_structure',
                'cosymlib.file_io',
                'cosymlib.symmetry',
                'cosymlib.symmetry.pointgroup',
                'cosymlib.tools',
                'cosymlib.simulation',
                'cosymlib.gui'],
      package_data={'': ['ideal_structures_center.yaml']},
      include_package_data=True,
      install_requires=['numpy', 'matplotlib', 'symgroupy', 'wfnsympy', 'PyYAML', 'huckelpy'],
      scripts=['scripts/cosym',
               'scripts/shape',
               'scripts/shape_map',
               'scripts/shape_classic'],
      ext_modules=[shape])



from cosymlib.shape import shp, shape_tools
import numpy as np
import hashlib


class Shape:
    def __init__(self, geometry):

        # Allow geometry or molecule to be imported instead of crude Cartesian coordinates
        try:
            self._coordinates = geometry.get_positions()
        except AttributeError:
            try:
                self._coordinates = geometry.geometry.get_positions()
            except AttributeError:
                self._coordinates = geometry

        self._coordinates = np.ascontiguousarray(self._coordinates)

        self._measures = {}
        self._structures = {}
        self._test_structures = {}
        self._path_deviation = {}
        self._gen_coord = {}

    # Function description
    def measure(self, label, central_atom=0, fix_permutation=False):
        hash = hashlib.md5('{}{}'.format(central_atom, label).encode()).hexdigest()
        if hash not in self._measures:
            if isinstance(label, str):
                reference_structure = shape_tools.get_test_structure(label, central_atom)
            else:
                reference_structure = np.array(label)

            if fix_permutation:
                self._measures[hash] = shp.cshm_fix(self._coordinates, reference_structure, central_atom)
            else:
                self._measures[hash] = shp.cshm(self._coordinates, reference_structure, central_atom)

        return self._measures[hash]

    # Function description
    def structure(self, label, central_atom=0, fix_permutation=False):
        hash = hashlib.md5('{}{}'.format(central_atom, label).encode()).hexdigest()
        if hash not in self._structures:
            if isinstance(label, str):
                reference_structure = shape_tools.get_test_structure(label, central_atom)
            else:
                reference_structure = np.array(label)

            if fix_permutation:
                self._structures[hash], self._measures[hash] = shp.poly_fix(self._coordinates, reference_structure,
                                                                            central_atom)
            else:
                self._structures[hash], self._measures[hash] = shp.poly(self._coordinates, reference_structure,
                                                                        central_atom)

        return self._structures[hash]

    def get_path_deviation(self, shape_label1, shape_label2, central_atom=0):

        hash = hashlib.md5('{}{}{}'.format(shape_label1, shape_label2, central_atom).encode()).hexdigest()
        if hash not in self._path_deviation:
            Sx = self.measure(shape_label1, central_atom)
            Sy = self.measure(shape_label2, central_atom)
            new_theta = np.arcsin(np.sqrt(Sx) / 10) + np.arcsin(np.sqrt(Sy) / 10)
            if isinstance(shape_label1, np.ndarray):
                structure_a = shape_label1
            else:
                structure_a = shape_tools.get_test_structure(shape_label1, central_atom=central_atom)
            theta = np.arcsin(np.sqrt(Shape(structure_a).measure(shape_label2, central_atom=len(structure_a))) / 10)
            self._path_deviation[hash] = ((new_theta / theta) - 1) * 100

        return self._path_deviation[hash]

    def get_generalized_coordinate(self, shape_label1, shape_label2, central_atom=0):

        hash = hashlib.md5('{}{}{}'.format(shape_label1, shape_label2, central_atom).encode()).hexdigest()
        if hash not in self._gen_coord:
            Sq = self.measure(shape_label1, central_atom)
            if isinstance(shape_label1, np.ndarray):
                structure_a = shape_label1
            else:
                structure_a = shape_tools.get_test_structure(shape_label1, central_atom=central_atom)
            theta = np.arcsin(np.sqrt(Shape(structure_a).measure(shape_label2, central_atom=len(structure_a))) / 10)
            self._gen_coord[hash] = round(100 * np.arcsin(np.sqrt(Sq) / 10) / theta, 1)

        return self._gen_coord[hash]






from cosymlib.shape import shape_tools, Shape
import numpy as np


def get_shape_map(shape_label1, shape_label2, num_points=20):

    path_structures = []
    if isinstance(shape_label1, np.ndarray):
        structure_a = shape_label1
    else:
        structure_a = shape_tools.get_test_structure(shape_label1, central_atom=1)
    structure_b = Shape(structure_a).structure(shape_label2, central_atom=len(structure_a))
    structure_b = np.concatenate((structure_b[1:], [structure_b[0]]))

    measure_label1 = []
    measure_label2 = []
    for i in range(0, num_points+1):
        structure_along_path = []
        for ida, atom in enumerate(structure_a):
            structure_along_path.append(atom + (structure_b[ida] - atom)*i/num_points)

        measure_label1.append(Shape(structure_along_path).measure(shape_label1, central_atom=len(structure_a)))
        measure_label2.append(Shape(structure_along_path).measure(shape_label2, central_atom=len(structure_a)))
        path_structures.append(structure_along_path)

    return measure_label1, measure_label2, path_structures



import os
import yaml
import numpy as np


ideal_structures = None


def get_test_structure(label, central_atom=0):

    global ideal_structures
    if ideal_structures is None:
        file_path = os.path.dirname(os.path.abspath(__file__)) + '/ideal_structures_center.yaml'
        with open(file_path, 'r') as stream:
            ideal_structures = yaml.load(stream, Loader=yaml.FullLoader)
    if central_atom == 0:
        reference_structure = ideal_structures[label][:-1]
    else:
        reference_structure = ideal_structures[label]

    return reference_structure


def get_structure_references(vertices):

    # print('Available reference structure with {} Vertices'.format(vertices))
    references_list = []
    for ref in shape_structure_references['{} Vertices'.format(vertices)]:
        references_list.append(ref[0])
    return references_list


def order_coordinates(coordinates, indices):
    indices = [i - 1 for i in indices]
    coordinates = np.array(coordinates).copy()
    coordinates[indices, :] = coordinates[indices[::-1], :]
    return coordinates


def get_shape_label(code, vertices):

    for label in shape_structure_references[str(vertices)+' Vertices']:
        if label[1] == code:
            return label[0]


def get_shape_label_info(vertices, old=False):
    txt = 'Available reference structures with {} Vertices:\n\n'.format(vertices)
    if old:
        txt += '{:<4} '.format('')
    txt += '{:10}  {:8}  {}\n\n'.format('Label', 'Sym', 'Info')
    for idl, labels in enumerate(shape_structure_references['{} Vertices'.format(vertices)]):
        if old:
            txt += '{:<4} '.format(idl+1)
        txt += '{:10}  {:8}  {}\n'.format(labels[0], labels[2], labels[3])
    return txt


shape_structure_references = {'2 Vertices': [['L-2', 1, 'Dinfh', 'Linear'],
                                             ['vT-2', 2, 'C2v', 'Divacant tetrahedron'],
                                             ['vOC-2', 3, 'C2v', 'Tetravacant octahedron']],
                              '3 Vertices': [['TP-3', 1, 'D3h', 'Trigonal'],
                                             ['vT-3', 2, 'C3v', 'Vacant tetrahedron'],
                                             ['fvOC-3', 3, 'C3v', 'fac-Trivacant octahedron'],
                                             ['mvOC-3', 4, 'C2v', 'mer-Trivacant octahedron']],
                              '4 Vertices': [['SP-4', 1, 'D4h', 'Square'],
                                             ['T-4', 2, 'Td', 'Tetrahedron'],
                                             ['SS-4', 3, 'C2v', 'Seesaw'],
                                             ['vTBPY-4', 4, 'C3v', 'Axially vacant trigonal bipyramid']],
                              '5 Vertices': [['PP-5', 1, 'D5h', 'Pentagon'],
                                             ['vOC-5', 2, 'C4v', 'Vacant octahedron'],
                                             ['TBPY-5', 3, 'D3h', 'Trigonal bipyramid'],
                                             ['SPY-5', 4, 'C4v', 'Spherical square pyramid'],
                                             ['JTBPY-5', 5, 'D3h', 'Johnson trigonal bipyramid J12']],
                              '6 Vertices': [['HP-6', 1, 'D6h', 'Hexagon'],
                                             ['PPY-6', 2, 'C5v', 'Pentagonal pyramid'],
                                             ['OC-6', 3, 'Oh', 'Octahedron'],
                                             ['TPR-6', 4, 'D3h', 'Trigonal prism'],
                                             ['JPPY-6', 5, 'C5v', 'Johnson pentagonal pyramid J2']],
                              '7 Vertices': [['HP-7', 1, 'D7h', 'Heptagon'],
                                             ['HPY-7', 2, 'C6v', 'Hexagonal pyramid'],
                                             ['PBPY-7', 3, 'D5h', 'Pentagonal bipyramid'],
                                             ['COC-7', 4, 'C3v', 'Capped octahedron'],
                                             ['CTPR-7', 5, 'C2v', 'Capped trigonal prism'],
                                             ['JPBPY-7', 6, 'D5h', 'Johnson pentagonal bipyramid J13'],
                                             ['JETPY-7', 7, 'C3v', 'Johnson elongated triangular pyramid J7']],
                              '8 Vertices': [['OP-8', 1, 'D8h', 'Octagon'],
                                             ['HPY-8', 2, 'C7v', 'Heptagonal pyramid'],
                                             ['HBPY-8', 3, 'D6h', 'Hexagonal bipyramid'],
                                             ['CU-8', 4, 'Oh', 'Cube'],
                                             ['SAPR-8', 5, 'D4d', 'Square antiprism'],
                                             ['TDD-8', 6, 'D2d', 'Triangular dodecahedron'],
                                             ['JGBF-8', 7, 'D2d', 'Johnson gyrobifastigium J26'],
                                             ['JETBPY-8', 8, 'D3h', 'Johnson elongated triangular bipyramid J14'],
                                             ['JBTPR-8', 9, 'C2v', 'Biaugmented trigonal prism J50'],
                                             ['BTPR-8', 10, 'C2v', 'Biaugmented trigonal prism'],
                                             ['JSD-8', 11, 'D2d', 'Snub diphenoid J84'],
                                             ['TT-8', 12, 'Td', 'Triakis tetrahedron'],
                                             ['ETBPY-8', 13, 'D3h', 'Elongated trigonal bipyramid']],
                              '9 Vertices': [['EP-9', 1, 'D9h', 'Enneagon'],
                                             ['OPY-9', 2, 'C8v', 'Octagonal pyramid'],
                                             ['HBPY-9', 3, 'D7h', 'Heptagonal bipyramid'],
                                             ['JTC-9', 4, 'C3v', 'Johnson triangular cupola J3'],
                                             ['JCCU-9', 5, 'C4v', 'Capped cube J8'],
                                             ['CCU-9', 6, 'C4v', 'Spherical-relaxed capped cube'],
                                             ['JCSAPR-9', 7, 'C4v', 'Capped square antiprism J10'],
                                             ['CSAPR-9', 8, 'C4v', 'Spherical capped square antiprism'],
                                             ['JTCTPR-9', 9, 'D3h', 'Tricapped trigonal prism J51'],
                                             ['TCTPR-9', 10, 'D3h', 'Spherical tricapped trigonal prism'],
                                             ['JTDIC-9', 11, 'C3v', 'Tridiminished icosahedron J63'],
                                             ['HH-9', 12, 'C2v', 'Hula-hoop'],
                                             ['MFF-9', 13, 'Cs', 'Muffin']],
                              '10 Vertices': [['DP-10', 1, 'D10h', 'Decagon'],
                                              ['EPY-10', 2, 'C9v', 'Enneagonal pyramid'],
                                              ['OBPY-10', 3, 'D8h', 'Octagonal bipyramid'],
                                              ['PPR-10', 4, 'D5h', 'Pentagonal prism'],
                                              ['PAPR-10', 5, 'D5d', 'Pentagonal antiprism'],
                                              ['JBCCU-10', 6, 'D4h', 'Bicapped cube J15'],
                                              ['JBCSAPR-10', 7, 'D4d', 'Bicapped square antiprism J17'],
                                              ['JMBIC-10', 8, 'C2v', 'Metabidiminished icosahedron J62'],
                                              ['JATDI-10', 9, 'C3v', 'Augmented tridiminished icosahedron J64'],
                                              ['JSPC-10', 10, 'C2v', 'Sphenocorona J87'],
                                              ['SDD-10', 11, 'D2', 'Staggered Dodecahedron (2:6:2)'],
                                              ['TD-10', 12, 'C2v', 'Tetradecahedron (2:6:2)'],
                                              ['HD-10', 13, 'D4h', 'Hexadecahedron (2:6:2) or (1:4:4:1)']],
                              '11 Vertices': [['HP-11', 1, 'D11h', 'Hendecagon'],
                                              ['DPY-11', 2, 'C10v', 'Decagonal pyramid'],
                                              ['EBPY-11', 3, 'D9h', 'Enneagonal bipyramid'],
                                              ['JCPPR-11', 4, 'C5v', 'Capped pentagonal prism J9'],
                                              ['JCPAPR-11', 5, 'C5v', 'Capped pentagonal antiprism J11'],
                                              ['JAPPR-11', 6, 'C2v', 'Augmented pentagonal prism J52'],
                                              ['JASPC-11', 7, 'Cs', 'Augmented sphenocorona J87']],
                              '12 Vertices': [['DP-12', 1, 'D12h', 'Dodecagon'],
                                              ['HPY-12', 2, 'C11v', 'Hendecagonal pyramid'],
                                              ['DBPY-12', 3, 'D10h', 'Decagonal bipyramid'],
                                              ['HPR-12', 4, 'D6h', 'Hexagonal prism'],
                                              ['HAPR-12', 5, 'D6d', 'Hexagonal antiprism'],
                                              ['TT-12', 6, 'Td', 'Truncated tetrahedron'],
                                              ['COC-12', 7, 'Oh', 'Cuboctahedron'],
                                              ['ACOC-12', 8, 'D3h', 'Anticuboctahedron J27'],
                                              ['IC-12', 9, 'Ih', 'Icosahedron'],
                                              ['JSC-12', 10, 'C4v', 'Johnson square cupola J4'],
                                              ['JEPBPY-12', 11, 'D6h', 'Johnson elongated pentagonal bipyramid J16'],
                                              ['JBAPPR-12', 12, 'C2v', 'Biaugmented pentagonal prism J53'],
                                              ['JSPMC-12', 13, 'Cs', 'Sphenomegacorona J88']],
                              '20 Vertices': [['DD-20', 1, 'Ih', 'Dodecahedron']],
                              '24 Vertices': [['TCU-24', 1, 'Oh', 'Truncated cube'],
                                              ['TOC-24', 2, 'Oh', 'Truncated octahedron']],
                              '48 Vertices': [['TCOC-48', 1, 'Oh', 'Truncated cuboctahedron']],
                              '60 Vertices': [['TRIC-60', 1, 'Ih', 'Truncated icosahedron (fullerene)']]}


import numpy as np
import hashlib
from symgroupy import Symgroupy


class Symgroup:

    def __init__(self, geometry):

        # Allow geometry or molecule to be imported instead of crude Cartesian coordinates
        try:
            self._coordinates = geometry.get_positions()
            self._symbols = geometry.get_symbols()
        except AttributeError:
            try:
                self._coordinates = geometry.geometry.get_positions()
                self._symbols = geometry.geometry.get_symbols()
            except AttributeError:
                self._coordinates = geometry
                self._symbols = None

        self._coordinates = np.ascontiguousarray(self._coordinates)

        self._results = {}

    # Function description
    def measure(self, label, central_atom=None, multi=1, symbols=True):

        hash = hashlib.md5('{}{}'.format(label, central_atom).encode()).hexdigest()
        if hash not in self._results:
            self._do_measure(label, central_atom=central_atom, multi=multi, symbols=symbols)

        return self._results[hash].csm

    # Function description
    def nearest_structure(self, label, central_atom=None, multi=1, symbols=True):

        hash = hashlib.md5('{}{}'.format(label, central_atom).encode()).hexdigest()
        if hash not in self._results:
            self._do_measure(label, central_atom=central_atom, multi=multi, symbols=symbols)

        return self._results[hash].nearest_structure

    def optimum_axis(self, label, central_atom=None, multi=1, symbols=True):

        hash = hashlib.md5('{}{}'.format(label, central_atom).encode()).hexdigest()
        if hash not in self._results:
            self._do_measure(label, central_atom=central_atom, multi=multi, symbols=symbols)

        return self._results[hash].optimum_axis

    def optimum_permutation(self, label, central_atom=None, multi=1, symbols=True):

        hash = hashlib.md5('{}{}'.format(label, central_atom).encode()).hexdigest()
        if hash not in self._results:
            self._do_measure(label, central_atom=central_atom, multi=multi, symbols=symbols)

        return self._results[hash].optimum_permutation

    def reference_axis(self, label, central_atom=None, multi=1, symbols=True):

        hash = hashlib.md5('{}{}'.format(label, central_atom).encode()).hexdigest()
        if hash not in self._results:
            self._do_measure(label, central_atom=central_atom, multi=multi, symbols=symbols)

        return self._results[hash].reference_axis

    def cms_multi(self, label, central_atom=None, multi=1, symbols=True):

        hash = hashlib.md5('{}{}'.format(label, central_atom).encode()).hexdigest()
        if hash not in self._results:
            self._do_measure(label, central_atom=central_atom, multi=multi, symbols=symbols)

        return self._results[hash].cms_multi

    def axis_multi(self, label, central_atom=None, multi=1, symbols=True):

        hash = hashlib.md5('{}{}'.format(label, central_atom).encode()).hexdigest()
        if hash not in self._results:
            self._do_measure(label, central_atom=central_atom, multi=multi, symbols=symbols)

        return self._results[hash].axis_multi

    def _do_measure(self, label, central_atom, multi, symbols=True):

        hash = hashlib.md5('{}{}'.format(label, central_atom).encode()).hexdigest()
        if not symbols:
            self._symbols = None
        self._results[hash] = Symgroupy(self._coordinates,
                                        group=label,
                                        labels=self._symbols,
                                        central_atom=central_atom,
                                        multi=multi)

    def results(self, label, central_atom=None, multi=1, symbols=True):

        hash = hashlib.md5('{}{}'.format(label, central_atom).encode()).hexdigest()
        if hash not in self._results:
            self._do_measure(label, central_atom=central_atom, multi=multi, symbols=symbols)

        return self._results[hash]